/* CATWALK - Test your logic
   Copyright (C) 2021 Valentin Moguerou
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULIAR PURPOSE. See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>  */

#include <stdio.h>
#include <string.h>

#include <catwalk/grid.h>
#include <catwalk/route.h>

#include "print.h"
#include "interact.h"

#define COMMAND_MAX_SIZE 100
#define STR_STARTS_WITH(s1, s2) (strncmp(s1, s2, strlen(s2)) == 0)

void print_help()
{
    printf(" | [l]eft, [r]ight     - move towards the specified direction\n");
    printf(" | [u]p, [d]own        - move towards the specified direction\n");
    printf(" | undo, z             - undo the last move\n");
    printf(" | [r]eset             - reset the grid\n");
    printf(" | help, ?             - show this help\n");
    printf(" | [s]how              - show the current grid\n");
    printf(" | exit, [q]uit        - exit the program\n");
}

void clear_buffer()
{
    int c = 0;
    while (c != '\n' && c != EOF)
    {
        c = getchar();
    }
}

int is_whitespace(char *str)
{
    if (*str == '\0')
    {
        return 1;
    }
    for (char *chr=str; *chr; chr++)
    {
        if (*chr != ' ' && *chr != '\t')
        {
            return 0;
        }
    }
    return 1;
}

void ltrim(char *dest, char *src)
{
    char *ch = src;
    while (*ch)
    {
        if (*ch != ' ' && *ch != '\t')
        {
            break;
        }
        ch++;
    }
    strcpy(dest, ch);
}

int read_from_stdin(char *str, int length)
{
    char *entry_position = NULL;

    if (fgets(str, length, stdin) != NULL)
    {
        entry_position = strchr(str, '\n');
        if (entry_position != NULL)
        {
            *entry_position = '\0';
        }
        else
        {
            clear_buffer();
        }
        return 1;
    }
    else
    {
        clear_buffer();
        return 0;
    }
}

void interact(int width)
{
    grid *gd = init_grid(width);

    printf("Welcome to Catwalk CLI. Type 'help' or '?' to see all available commands.\n");
    char prompt[COMMAND_MAX_SIZE] = "";
    char buffer[COMMAND_MAX_SIZE] = "";

    refresh_grid(gd);
    init_player_route(gd);

    for(;;)
    {
        if (is_whitespace(prompt))
        {
            printf("catwalk> ");
            read_from_stdin(prompt, COMMAND_MAX_SIZE);
            ltrim(buffer, prompt);
            strcpy(prompt, buffer);
        }
        else
        {
            if STR_STARTS_WITH(prompt, "help")
            {
                print_help();
                strcpy(buffer, prompt+strlen("help"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "?")
            {
                print_help();
                strcpy(buffer, prompt+strlen("?"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "show")
            {
                print_interactive_grid(gd);
                strcpy(buffer, prompt+strlen("show"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "reset")
            {
                if (gd->player_route->first->next == NULL)
                {
                    printf("Play first!\n");
                }
                else
                {
                    delete_element(gd->player_route->first->next);
                    gd->player_route->first->next = NULL;
                    gd->player_route->last = gd->player_route->first;
                }
                strcpy(buffer, prompt+strlen("reset"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "route")
            {
                print_route(gd->player_route);
                strcpy(buffer, prompt+strlen("route"));
                ltrim(prompt, buffer);
            }
            else if (STR_STARTS_WITH(prompt, "exit") || STR_STARTS_WITH(prompt, "quit") || STR_STARTS_WITH(prompt, "q"))
            {
                break;
            }
            else if STR_STARTS_WITH(prompt, "undo")
            {
                if (gd->player_route->first->next == NULL)
                {
                    printf("Can't undo! (single element)\n");
                }
                else
                {
                    pop_back(gd->player_route);
                }
                strcpy(buffer, prompt+strlen("z"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "left")
            {
                switch (move_left(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("left"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "right")
            {
                switch (move_right(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("right"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "up")
            {
                switch (move_up(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("up"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "down")
            {
                switch (move_down(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("down"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "s")
            {
                print_interactive_grid(gd);
                strcpy(buffer, prompt+strlen("s"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "l")
            {
                switch (move_left(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("l"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "r")
            {
                switch (move_right(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("r"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "u")
            {
                switch (move_up(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("u"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "d")
            {
                if (gd->player_route->first->next == NULL)
                {
                    printf("Can't undo! (single element)\n");
                }
                else
                {
                    pop_back(gd->player_route);
                }
                strcpy(buffer, prompt+strlen("undo"));
                ltrim(prompt, buffer);
            }
            else if STR_STARTS_WITH(prompt, "z")
            {
                switch (move_down(gd->player_route))
                {
                    case 1: puts("Cannot move: border of the grid"); break;
                    case 2: puts("Cannot move: already went there"); break;
                }
                strcpy(buffer, prompt+strlen("d"));
                ltrim(prompt, buffer);
            }
            else
            {
                printf("Unrecognized command : '%s'!\n", prompt);
                prompt[0] = '\0';
            }
        }


        if (gd->player_route->last->x == gd->end[0] && gd->player_route->last->y == gd->end[1])
        {
            print_interactive_grid(gd);
            if (verify(gd))
            {
                printf("You won!\n");
            }
            else
            {
                printf("You lost!\n");
            }

            random_start(gd);
            refresh_grid(gd);
            reset_player_route(gd);
        }
    }

    delete_grid(gd);
}
